/*
 * Common clock functionality.
 *
 * Copyright (C) 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#pragma once

#ifdef __cplusplus

#include <l4/devices/clocks.h>
#include <l4/devices/hw_register_block.h>
#include <l4/sys/types.h>
#include <stdint.h>

/* Forward declaration. */

class Clock_base;



/* Register access type. */

class Cpm_regs
{
protected:
  Hw::Register_block<32> _regs;
  Clock_base **_clocks;

public:
  explicit Cpm_regs(l4_addr_t addr, Clock_base *clocks[]);

  // Utility methods.

  uint32_t get_field(uint32_t reg, uint32_t mask, uint8_t shift);
  void set_field(uint32_t reg, uint32_t mask, uint8_t shift, uint32_t value);

  Clock_base *get_clock(int num);
};



// Register field abstraction.

class Field
{
  uint32_t reg;
  uint32_t mask;
  uint8_t bit;
  bool defined;
  uint32_t _asserted = 0, _deasserted = 0;

public:
  explicit Field()
  : defined(false)
  {
  }

  explicit Field(uint32_t reg, uint32_t mask, uint32_t bit,
                 bool inverted = false)
  : reg(reg), mask(mask), bit(bit), defined(true)
  {
    if (inverted)
      _deasserted = mask;
    else
      _asserted = mask;
  }

  uint32_t get_field(Cpm_regs &regs);
  void set_field(Cpm_regs &regs, uint32_t value);

  uint32_t get_asserted() { return _asserted; }
  uint32_t get_deasserted() { return _deasserted; }

  bool is_defined() { return defined; }
  uint32_t get_limit() { return mask; }

  // Undefined field object.

  static Field undefined;
};



// Clock sources.

class Mux
{
  int _num_inputs;
  enum Clock_identifiers *_inputs, _input;

public:
  explicit Mux(int num_inputs, enum Clock_identifiers inputs[])
  : _num_inputs(num_inputs), _inputs(inputs)
  {
  }

  explicit Mux(enum Clock_identifiers input)
  : _num_inputs(1), _inputs(&_input)
  {
    _input = input;
  }

  explicit Mux()
  : _num_inputs(0), _inputs(NULL)
  {
  }

  int get_number() { return _num_inputs; }
  enum Clock_identifiers get_input(int num);
};



// Controllable clock source.

class Source
{
  Mux _inputs;
  Field _source;

public:
  explicit Source(Mux inputs, Field source)
  : _inputs(inputs), _source(source)
  {
  }

  explicit Source(Mux inputs)
  : _inputs(inputs)
  {
  }

  explicit Source()
  {
  }

  int get_number() { return _inputs.get_number(); }
  enum Clock_identifiers get_input(int num) { return _inputs.get_input(num); }

  // Clock source.

  uint8_t get_source(Cpm_regs &regs);
  void set_source(Cpm_regs &regs, uint8_t source);
  enum Clock_identifiers get_source_clock(Cpm_regs &regs);
  void set_source_clock(Cpm_regs &regs, enum Clock_identifiers clock);

  // Clock source frequency.

  uint64_t get_frequency(Cpm_regs &regs);
};



// Common clock control.

class Control_base
{
public:
  virtual ~Control_base();

  virtual void change_disable(Cpm_regs &regs);
  virtual void change_enable(Cpm_regs &regs);

  virtual void wait_busy(Cpm_regs &regs) = 0;
  virtual int have_clock(Cpm_regs &regs) = 0;
  virtual void start_clock(Cpm_regs &regs) = 0;
  virtual void stop_clock(Cpm_regs &regs) = 0;
};



// Clock control.

class Control : public Control_base
{
  Field _gate, _change_enable, _busy;

public:
  explicit Control(Field gate,
                   Field change_enable = Field::undefined,
                   Field busy = Field::undefined)
  : _gate(gate), _change_enable(change_enable), _busy(busy)
  {
  }

  explicit Control()
  : _gate(Field::undefined), _change_enable(Field::undefined),
    _busy(Field::undefined)
  {
  }

  // Clock control.

  void change_disable(Cpm_regs &regs);
  void change_enable(Cpm_regs &regs);

  void wait_busy(Cpm_regs &regs);
  int have_clock(Cpm_regs &regs);
  void start_clock(Cpm_regs &regs);
  void stop_clock(Cpm_regs &regs);

  // Undefined control object.

  static Control undefined;
};



// PLL control.

class Control_pll : public Control_base
{
  Field _enable, _stable, _bypass;

  // Frequency change sequence state.

  bool _enabled = false;

  // PLL_specific control.

  int have_pll(Cpm_regs &regs);
  int pll_enabled(Cpm_regs &regs);

public:
  explicit Control_pll(Field enable, Field stable, Field bypass)
  : _enable(enable), _stable(stable), _bypass(bypass)
  {
  }

  // Clock control.

  int pll_bypassed(Cpm_regs &regs);

  void pll_bypass(Cpm_regs &regs);
  void pll_engage(Cpm_regs &regs);

  void change_disable(Cpm_regs &regs);
  void change_enable(Cpm_regs &regs);

  void wait_busy(Cpm_regs &regs);
  int have_clock(Cpm_regs &regs);
  void start_clock(Cpm_regs &regs);
  void stop_clock(Cpm_regs &regs);
};



// Frequency transformation.

class Divider_base
{
public:
  virtual ~Divider_base();

  // Output frequency.

  virtual uint64_t get_frequency(Cpm_regs &regs, uint64_t source_frequency) = 0;
  virtual int set_frequency(Cpm_regs &regs, uint64_t source_frequency, uint64_t frequency) = 0;

  // Other operations.

  virtual int get_parameters(Cpm_regs &regs, uint32_t parameters[]) = 0;

  virtual int set_parameters(Cpm_regs &regs, int num_parameters, uint32_t parameters[]) = 0;
};



// Simple divider for regular clocks.

class Divider : public Divider_base
{
  Field _divider;

public:
  explicit Divider(Field divider)
  : _divider(divider)
  {
  }

  explicit Divider()
  : _divider(Field::undefined)
  {
  }

  // Clock divider.

  uint32_t get_divider(Cpm_regs &regs);
  void set_divider(Cpm_regs &regs, uint32_t divider);

  // Output frequency.

  uint64_t get_frequency(Cpm_regs &regs, uint64_t source_frequency);
  int set_frequency(Cpm_regs &regs, uint64_t source_frequency, uint64_t frequency);

  // Other operations.

  int get_parameters(Cpm_regs &regs, uint32_t parameters[]);
  int set_parameters(Cpm_regs &regs, int num_parameters, uint32_t parameters[]);
};



// Divider for PLLs.

class Divider_pll : public Divider_base
{
  Field _multiplier, _input_divider, _output_divider0, _output_divider1;
  double _intermediate_min, _intermediate_max;
  bool _adjust_by_one;

  // General frequency modifiers.

  uint32_t get_multiplier(Cpm_regs &regs);
  void set_multiplier(Cpm_regs &regs, uint32_t multiplier);
  uint32_t get_input_divider(Cpm_regs &regs);
  void set_input_divider(Cpm_regs &regs, uint32_t divider);
  uint32_t get_output_divider(Cpm_regs &regs);
  void set_output_divider(Cpm_regs &regs, uint32_t divider);

public:

  // Double output divider constructor.

  explicit Divider_pll(Field multiplier, Field input_divider,
                       Field output_divider0, Field output_divider1,
                       double intermediate_min, double intermediate_max,
                       bool adjust_by_one)
  : _multiplier(multiplier), _input_divider(input_divider),
    _output_divider0(output_divider0), _output_divider1(output_divider1),
    _intermediate_min(intermediate_min), _intermediate_max(intermediate_max),
    _adjust_by_one(adjust_by_one)
  {
  }

  // Single output divider constructor.

  explicit Divider_pll(Field multiplier, Field input_divider,
                       Field output_divider,
                       double intermediate_min, double intermediate_max,
                       bool adjust_by_one)
  : _multiplier(multiplier), _input_divider(input_divider),
    _output_divider0(output_divider), _output_divider1(Field::undefined),
    _intermediate_min(intermediate_min), _intermediate_max(intermediate_max),
    _adjust_by_one(adjust_by_one)
  {
  }

  // Output frequency.

  uint64_t get_frequency(Cpm_regs &regs, uint64_t source_frequency);
  int set_frequency(Cpm_regs &regs, uint64_t source_frequency, uint64_t frequency);

  // Other operations.

  int get_parameters(Cpm_regs &regs, uint32_t parameters[]);
  int set_parameters(Cpm_regs &regs, int num_parameters, uint32_t parameters[]);
};



// Divider for I2S clocks.

class Divider_i2s : public Divider_base
{
  Field _multiplier, _divider_N, _divider_D, _auto_N, _auto_D;

  // General frequency modifiers.

  uint32_t get_multiplier(Cpm_regs &regs);
  uint32_t get_divider_N(Cpm_regs &regs);
  uint32_t get_divider_D(Cpm_regs &regs);

public:
  explicit Divider_i2s(Field multiplier, Field divider_N, Field divider_D,
                       Field auto_N, Field auto_D)
  : _multiplier(multiplier), _divider_N(divider_N), _divider_D(divider_D),
    _auto_N(auto_N), _auto_D(auto_D)
  {
  }

  // Output frequency.

  uint64_t get_frequency(Cpm_regs &regs, uint64_t source_frequency);
  int set_frequency(Cpm_regs &regs, uint64_t source_frequency, uint64_t frequency);

  // Other operations.

  int get_parameters(Cpm_regs &regs, uint32_t parameters[]);

  int set_parameters(Cpm_regs &regs, int num_parameters, uint32_t parameters[]);
};



// Common clock abstraction.

class Clock_base
{
public:
  virtual ~Clock_base();

  virtual const char *clock_type() { return "unset"; }

  // Clock control.

  virtual int have_clock(Cpm_regs &regs) = 0;
  virtual void start_clock(Cpm_regs &regs) = 0;
  virtual void stop_clock(Cpm_regs &regs) = 0;

  // Output frequency.

  virtual uint64_t get_frequency(Cpm_regs &regs) = 0;
};



// Null (absent or undefined) clock abstraction.

class Clock_null : public Clock_base
{
public:
  const char *clock_type() { return "null"; }

  // Clock control.

  int have_clock(Cpm_regs &regs);
  void start_clock(Cpm_regs &regs);
  void stop_clock(Cpm_regs &regs);

  // Output frequency.

  uint64_t get_frequency(Cpm_regs &regs);
};



// Passive (root or input) clock without any source of its own.

class Clock_passive : public Clock_base
{
protected:
  uint64_t _frequency;

public:
  explicit Clock_passive(uint64_t frequency)
  : _frequency(frequency)
  {
  }

  const char *clock_type() { return "passive"; }

  // Clock control.

  virtual int have_clock(Cpm_regs &regs);
  virtual void start_clock(Cpm_regs &regs);
  virtual void stop_clock(Cpm_regs &regs);

  // Output frequency.

  uint64_t get_frequency(Cpm_regs &regs);
};



class Clock_controlled : public Clock_base
{
protected:
  virtual Control_base &_get_control() = 0;

public:

  // Clock control.

  virtual int have_clock(Cpm_regs &regs);
  virtual void start_clock(Cpm_regs &regs);
  virtual void stop_clock(Cpm_regs &regs);
};



// An actively managed clock with source.

class Clock_active : public Clock_controlled
{
protected:
  Source _source;

public:
  explicit Clock_active(Source source)
  : _source(source)
  {
  }

  virtual ~Clock_active();

  // Clock source.

  virtual uint8_t get_source(Cpm_regs &regs);
  virtual void set_source(Cpm_regs &regs, uint8_t source);
  enum Clock_identifiers get_source_clock(Cpm_regs &regs);
  void set_source_clock(Cpm_regs &regs, enum Clock_identifiers clock);

  // Clock source frequency.

  virtual uint64_t get_source_frequency(Cpm_regs &regs);

  // Output frequency.

  virtual uint64_t get_frequency(Cpm_regs &regs);
};



// Divided clock interface.

class Clock_divided_base : public Clock_active
{
protected:
  virtual Divider_base &_get_divider() = 0;

public:
  explicit Clock_divided_base(Source source)
  : Clock_active(source)
  {
  }

  virtual ~Clock_divided_base();

  virtual int get_parameters(Cpm_regs &regs, uint32_t parameters[]);
  virtual int set_parameters(Cpm_regs &regs, int num_parameters, uint32_t parameters[]);

  // Output frequency.

  virtual uint64_t get_frequency(Cpm_regs &regs);
  virtual int set_frequency(Cpm_regs &regs, uint64_t frequency);
};



// PLL description.

class Pll : public Clock_divided_base
{
  Control_pll _control;
  Divider_pll _divider;

  virtual Control_base &_get_control() { return _control; }
  virtual Divider_base &_get_divider() { return _divider; }

public:
  explicit Pll(Source source, Control_pll control, Divider_pll divider)
  : Clock_divided_base(source), _control(control), _divider(divider)
  {
  }

  virtual ~Pll();

  const char *clock_type() { return "pll"; }

  // Output frequency.

  uint64_t get_frequency(Cpm_regs &regs);
  int set_frequency(Cpm_regs &regs, uint64_t frequency);
};



// Plain clock description.

class Clock : public Clock_active
{
  Control _control;

  virtual Control_base &_get_control() { return _control; }

public:
  explicit Clock(Source source, Control control)
  : Clock_active(source), _control(control)
  {
  }

  explicit Clock(Source source)
  : Clock_active(source)
  {
  }

  const char *clock_type() { return "clock"; }
};



// Divided clock description.

class Clock_divided : public Clock_divided_base
{
  Control _control;
  Divider _divider;

  virtual Control_base &_get_control() { return _control; }
  virtual Divider_base &_get_divider() { return _divider; }

public:
  explicit Clock_divided(Source source, Control control, Divider divider)
  : Clock_divided_base(source), _control(control), _divider(divider)
  {
  }

  explicit Clock_divided(Source source, Divider divider)
  : Clock_divided_base(source), _control(Control::undefined), _divider(divider)
  {
  }

  const char *clock_type() { return "divided"; }
};



// I2S clock description.

class Clock_divided_i2s : public Clock_divided_base
{
  Control _control;
  Divider_i2s _divider;

  virtual Control_base &_get_control() { return _control; }
  virtual Divider_base &_get_divider() { return _divider; }

public:
  explicit Clock_divided_i2s(Source source, Control control, Divider_i2s divider)
  : Clock_divided_base(source), _control(control), _divider(divider)
  {
  }

  const char *clock_type() { return "i2s"; }
};

#endif /* __cplusplus */
