/*
 * Perform SPI communication using GPIO operations.
 *
 * Copyright (C) 2018, 2020, 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/spi-gpio.h>
#include <time.h>



Spi_gpio::Spi_gpio(Hw::Gpio_chip *clock_device, int clock_pin,
                   Hw::Gpio_chip *data_device, int data_pin,
                   Hw::Gpio_chip *enable_device, int enable_pin,
                   Hw::Gpio_chip *control_device, int control_pin)
: _clock_device(clock_device),
  _clock_pin(clock_pin),
  _data_device(data_device),
  _data_pin(data_pin),
  _enable_device(enable_device),
  _enable_pin(enable_pin),
  _control_device(control_device),
  _control_pin(control_pin)
{
  _clock_device->setup(_clock_pin, Hw::Gpio_chip::Output, 1);
  _data_device->setup(_data_pin, Hw::Gpio_chip::Output, 0);
  _enable_device->setup(_enable_pin, Hw::Gpio_chip::Output, 1);
  _control_device->setup(_control_pin, Hw::Gpio_chip::Output, 0);
}

Spi_gpio::Spi_gpio(uint64_t frequency,
                   Hw::Gpio_chip *clock_device, int clock_pin,
                   Hw::Gpio_chip *data_device, int data_pin,
                   Hw::Gpio_chip *enable_device, int enable_pin,
                   Hw::Gpio_chip *control_device, int control_pin)
: Spi_gpio(clock_device, clock_pin, data_device, data_pin, enable_device,
           enable_pin, control_device, control_pin)
{
  _frequency = frequency;
}

/* Send a byte sequence. */

uint32_t Spi_gpio::send(uint32_t bytes, const uint8_t data[])
{
  return send_dc(bytes, data, NULL);
}

/* Send a byte sequence with accompanying data/command indicators. */

uint32_t Spi_gpio::send_dc(uint32_t bytes, const uint8_t data[],
                           const int dc[])
{
  struct timespec ts;
  uint8_t mask;
  int bit;
  uint32_t byte;

  if (_frequency)
  {
    ts.tv_sec = 0;
    ts.tv_nsec = 1000000000 / _frequency;
  }

  /* Initialise pin levels. */

  _enable_device->set(_enable_pin, 1);
  _clock_device->set(_clock_pin, 1);
  _data_device->set(_data_pin, 0);

  /* Enter the transmission state. */

  _enable_device->set(_enable_pin, 0);

  /* Clock data using the clock and data outputs. */

  for (byte = 0; byte < bytes; byte++)
  {
    mask = 0x80;

    for (bit = 0; bit < 8; bit++)
    {
      /* NOTE: Data presented on falling clock level and sampled on rising clock
               level. This is SPI mode 3, or 0 given that the enable level is
               driven low immediately before the first bit is presented. */

      _clock_device->set(_clock_pin, 0);
      _data_device->set(_data_pin, data[byte] & mask ? 1 : 0);

      if ((_control_device != NULL) && (_control_pin >= 0) && (dc != NULL))
        _control_device->set(_control_pin, dc[byte] ? 1 : 0);

      if (_frequency)
        nanosleep(&ts, NULL);

      _clock_device->set(_clock_pin, 1);

      if (_frequency)
        nanosleep(&ts, NULL);

      mask >>= 1;
    }
  }

  _enable_device->set(_enable_pin, 1);

  return bytes;
}

/* Send a number of units, each holding a value in a big endian arrangement
   limited to the given character size, with any bit immediately above the
   character portion of the unit indicating the data/command level. */

uint32_t Spi_gpio::send_units(uint32_t bytes, const uint8_t data[],
                              uint8_t unit_size, uint8_t char_size)
{
  /* Obtain the data/command values for each unit. */

  uint32_t chars = bytes / unit_size;
  int dc[chars];

  /* Obtain the actual character bytes to be sent. */

  int char_unit_size = ((char_size ? char_size - 1 : 0) / 8) + 1;
  uint8_t char_data[char_unit_size * chars];

  /* Traverse the byte sequence, extracting data/command bits for each unit. */

  for (uint32_t offset = 0, unit = 0, char_offset = 0; unit < chars;
       offset += unit_size, unit++, char_offset += char_unit_size)
  {
    /* Obtain the unit value. */

    uint32_t value = 0;

    for (uint8_t byte = 0; byte < unit_size; byte++)
      value = (value << 8) | data[offset + byte];

    /* The unit size must be greater than the character size for data/command
       bits to be present. */

    if (unit_size * 8 <= char_size)
      dc[unit] = 0;

    /* Extract the data/command level. */

    else
      dc[unit] = value & (1 << char_size) ? 1 : 0;

    /* Obtain the actual character data from the input data, discarding the most
       significant bytes beyond the character and masking out any remaining
       superfluous bits, thus obtaining a whole number of bytes. */

    value = value & ((1 << char_size) - 1);

    for (uint8_t byte = char_unit_size; byte != 0; byte--)
    {
      char_data[char_offset + byte - 1] = value & 0xff;
      value >>= 8;
    }
  }

  return send_dc(chars, char_data, dc);
}



/* C language interface. */

void *spi_gpio_get_channel(uint64_t frequency,
                           void *clock_chip, int clock_pin,
                           void *data_chip, int data_pin,
                           void *enable_chip, int enable_pin,
                           void *control_chip, int control_pin)
{
  return (void *) new Spi_gpio(frequency,
                               reinterpret_cast<Hw::Gpio_chip *>(clock_chip), clock_pin,
                               reinterpret_cast<Hw::Gpio_chip *>(data_chip), data_pin,
                               reinterpret_cast<Hw::Gpio_chip *>(enable_chip), enable_pin,
                               reinterpret_cast<Hw::Gpio_chip *>(control_chip), control_pin);
}

uint32_t spi_gpio_send(void *channel, uint32_t bytes, const uint8_t data[])
{
  return static_cast<Spi_gpio *>(channel)->send(bytes, data);
}

uint32_t spi_gpio_send_dc(void *channel, uint32_t bytes, const uint8_t data[],
                          const int dc[])
{
  return static_cast<Spi_gpio *>(channel)->send_dc(bytes, data, dc);
}

uint32_t spi_gpio_send_units(void *channel, uint32_t bytes, const uint8_t data[],
                             uint8_t unit_size, uint8_t char_size)
{
  return static_cast<Spi_gpio *>(channel)->send_units(bytes, data, unit_size, char_size);
}
