/*
 * Perform SPI communication using a suitable abstraction augmented with
 * explicit manipulation of a control signal.
 *
 * Copyright (C) 2023 Paul Boddie <paul@boddie.org.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA
 */

#include <l4/devices/spi-hybrid.h>



Spi_hybrid::Spi_hybrid(Spi_channel_base *channel,
                       Hw::Gpio_chip *control_device, int control_pin,
                       int control_alt_func)
: _channel(channel),
  _control_device(control_device),
  _control_pin(control_pin),
  _control_alt_func(control_alt_func)
{
}

Spi_channel_base *Spi_hybrid::get_channel()
{
  return _channel;
}

void Spi_hybrid::acquire_control(bool asserted)
{
  _control_device->setup(_control_pin, Hw::Gpio_chip::Output, asserted ? 1 : 0);
}

void Spi_hybrid::release_control()
{
  if (_control_alt_func >= 0)
    _control_device->config_pad(_control_pin, Hw::Gpio_chip::Function_alt, _control_alt_func);
  else
    _control_device->setup(_control_pin, Hw::Gpio_chip::Input, 0);
}

/* Send a byte sequence. */

uint32_t Spi_hybrid::send(uint32_t bytes, const uint8_t data[])
{
  return _channel->send(bytes, data);
}
  
/* Send a byte sequence with control information. */

uint32_t Spi_hybrid::send_dc(uint32_t bytes, const uint8_t data[], const int dc[])
{
  return _channel->send_dc(bytes, data, dc);
}
  
/* Send a sequence of units having the given character size. */

uint32_t Spi_hybrid::send_units(uint32_t bytes, const uint8_t data[],
                                uint8_t unit_size, uint8_t char_size)
{
  return _channel->send_units(bytes, data, unit_size, char_size);
}
  


/* C language interface. */

void *spi_hybrid_get_channel(void *channel, void *control_chip, int control_pin,
                             int control_alt_func)
{
  return (void *) new Spi_hybrid(reinterpret_cast<Spi_channel_base *>(channel),
                                 reinterpret_cast<Hw::Gpio_chip *>(control_chip),
                                 control_pin, control_alt_func);
}

void *spi_hybrid_get_raw_channel(void *channel)
{
  return static_cast<Spi_hybrid *>(channel)->get_channel();
}

void spi_hybrid_acquire_control(void *channel, int level)
{
  static_cast<Spi_hybrid *>(channel)->acquire_control(level);
}

void spi_hybrid_release_control(void *channel)
{
  static_cast<Spi_hybrid *>(channel)->release_control();
}

uint32_t spi_hybrid_send(void *channel, uint32_t bytes, const uint8_t data[])
{
  return static_cast<Spi_hybrid *>(channel)->send(bytes, data);
}

uint32_t spi_hybrid_send_dc(void *channel, uint32_t bytes, const uint8_t data[],
                            const int dc[])
{
  return static_cast<Spi_hybrid *>(channel)->send_dc(bytes, data, dc);
}

uint32_t spi_hybrid_send_units(void *channel, uint32_t bytes, const uint8_t data[],
                               uint8_t unit_size, uint8_t char_size)
{
  return static_cast<Spi_hybrid *>(channel)->send_units(bytes, data, unit_size, char_size);
}
